"""Support filters for Peewee ORM."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional
from warnings import warn

if TYPE_CHECKING:
    from peewee import Field


def get_model_field_by_name(handler, name: str, *, stacklevel=5) -> Optional[Field]:
    """Get model field by name."""
    fields = handler.meta.model._meta.fields
    candidate = fields.get(name)
    if candidate:
        return candidate

    for field in fields.values():
        if field.column_name == name:
            return field

    warn(
        f"{handler.__qualname__} {handler.meta.model} has no field {name}",
        category=RuntimeWarning,
        stacklevel=stacklevel,
    )
    return None



"""REST helpers for Muffin Framework."""

from contextlib import suppress

# Default query params
LIMIT_PARAM = "limit"
OFFSET_PARAM = "offset"


from .api import API
from .errors import APIError
from .handler import RESTHandler

# Just an alias to support legacy style
Api = API


__all__ = (
    "API",
    "Api",
    "RESTHandler",
    "APIError",
    "PWRESTHandler",
    "PWFilter",
    "PWFilters",
    "PWSort",
    "PWSorting",
    "SARESTHandler",
    "SAFilter",
    "SAFilters",
    "SASort",
    "SASorting",
    "MongoRESTHandler",
    "MongoFilter",
    "MongoFilters",
    "MongoSort",
    "MongoSorting",
)

# Support Peewee ORM
with suppress(ImportError):
    from .peewee import PWRESTHandler
    from .peewee.filters import PWFilter, PWFilters
    from .peewee.sorting import PWSort, PWSorting


# Support SQLAlchemy ORM
with suppress(ImportError):
    from .sqlalchemy import SARESTHandler
    from .sqlalchemy.filters import SAFilter, SAFilters
    from .sqlalchemy.sorting import SASort, SASorting


# Support Mongo ORM
with suppress(ImportError):
    from .mongo import MongoRESTHandler
    from .mongo.filters import MongoFilter, MongoFilters
    from .mongo.sorting import MongoSort, MongoSorting

# ruff: noqa: E402



import peewee as pw
from marshmallow_peewee import ModelSchema
from peewee_aio import Manager

from muffin_rest.options import RESTOptions

from .filters import PWFilters
from .sorting import PWSorting


class PWRESTOptions(RESTOptions):
    """Support Peewee."""

    # Base filters class
    filters_cls: type[PWFilters] = PWFilters

    # Base sorting class
    sorting_cls: type[PWSorting] = PWSorting

    Schema: type[ModelSchema]

    # Schema auto generation params
    schema_base: type[ModelSchema] = ModelSchema

    base_property: str = "model"

    model: type[pw.Model]
    model_pk: pw.Field

    manager: Manager

    # Recursive delete
    delete_recursive = False

    def setup(self, cls):
        """Prepare meta options."""
        meta = self.model._meta  # type: ignore[]
        self.name = self.name or meta.table_name.lower()
        self.model_pk = getattr(self, "model_pk", None) or meta.primary_key
        manager = getattr(self, "manager", getattr(self.model, "_manager", None))
        if manager is None:
            raise RuntimeError("Peewee-AIO ORM Manager is not available")

        self.manager = manager

        super().setup(cls)

    def setup_schema_meta(self, _):
        """Prepare a schema."""
        return type(
            "Meta",
            (object,),
            dict(
                {"unknown": self.schema_unknown, "model": self.model},
                **self.schema_meta,
            ),
        )



"""Support openapi."""

from __future__ import annotations

from typing import TYPE_CHECKING

from muffin_rest.openapi import OpenAPIMixin

if TYPE_CHECKING:
    from apispec import APISpec
    from http_router.routes import Route

    from .options import PWRESTOptions


class PeeweeOpenAPIMixin(OpenAPIMixin):
    """Render openapi."""

    meta: PWRESTOptions

    @classmethod
    def openapi(cls, route: Route, spec: APISpec, tags: dict) -> dict:
        """Get openapi specs for the endpoint."""
        operations = super(PeeweeOpenAPIMixin, cls).openapi(route, spec, tags)
        is_resource_route = getattr(route, "params", {}).get(cls.meta.name_id)
        if not is_resource_route and "delete" in operations:
            operations["delete"].setdefault("parameters", [])
            operations["delete"]["requestBody"] = {
                "required": True,
                "content": {
                    "application/json": {
                        "schema": {"type": "array", "items": {"type": "string"}},
                    },
                },
            }
        return operations



"""Support filters for SQLAlchemy ORM."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Union, cast

from sqlalchemy import Column

from muffin_rest.filters import Filter, Filters

if TYPE_CHECKING:
    from muffin_rest.types import TFilterValue

    from .types import TVCollection


class SAFilter(Filter):
    """Custom filter for sqlalchemy."""

    operators: ClassVar = dict(Filter.operators)
    operators["$between"] = lambda c, v: c.between(*v)
    operators["$ends"] = lambda c, v: c.endswith(v)
    operators["$ilike"] = lambda c, v: c.ilike(v)
    operators["$in"] = lambda c, v: c.in_(v)
    operators["$like"] = lambda c, v: c.like(v)
    operators["$match"] = lambda c, v: c.match(v)
    operators["$nin"] = lambda c, v: ~c.in_(v)
    operators["$notilike"] = lambda c, v: c.notilike(v)
    operators["$notlike"] = lambda c, v: c.notlike(v)
    operators["$starts"] = lambda c, v: c.startswith(v)

    list_ops = (*Filter.list_ops, "$between")

    async def filter(self, collection: TVCollection, *ops: TFilterValue) -> TVCollection:
        """Apply the filters to SQLAlchemy Select."""
        column = self.field
        if ops and column is not None:
            return self.query(collection, column, *ops)

        return collection

    def query(self, select: TVCollection, column: Column, *ops: TFilterValue) -> TVCollection:
        """Filter a select."""
        return select.where(*[op(column, val) for op, val in ops])


class SAFilters(Filters):
    """Bind SAfilter class."""

    MUTATE_CLASS = SAFilter

    def convert(self, obj: Union[str, Column, SAFilter], **meta):
        """Convert params to filters."""
        from . import SARESTHandler

        handler = cast(SARESTHandler, self.handler)

        if isinstance(obj, SAFilter):
            if obj.field is None:
                obj.field = handler.meta.table.c.get(obj.name)
            return obj

        if isinstance(obj, Column):
            name = obj.name
            field: Any = obj

        else:
            name = obj
            field = meta.pop("field", None) or name
            if isinstance(field, str):
                field = handler.meta.table.c.get(field)

        schema_field = meta.pop("schema_field", None)
        if schema_field is None and field is not None:
            schema_field = handler.meta.Schema._declared_fields.get(field.name)

        return self.MUTATE_CLASS(name, field=field, schema_field=schema_field, **meta)



import abc
from time import time
from typing import Any


class RateLimiter(abc.ABC):
    """Rate limiter."""

    def __init__(self, limit: int, *, period: int = 60, **opts):
        """Initialize the rate limiter.

        Args:
            limit (int): The limit of requests.
            period (int): The period of time in seconds.
        """
        self.limit = limit
        self.period = period

    @abc.abstractmethod
    async def check(self, key: str) -> bool:
        """Check the request."""
        raise NotImplementedError


class MemoryRateLimiter(RateLimiter):
    """Memory rate limiter. Do not use in production."""

    def __init__(self, limit: int, **opts):
        super().__init__(limit, **opts)
        self.storage: dict[Any, tuple[float, int]] = {}

    async def check(self, key: str) -> bool:
        """Check the request."""
        now = time()
        storage = self.storage

        if key not in storage:
            storage[key] = (now, 1)
            return True

        last, count = storage[key]
        if now - last > self.period:
            storage[key] = (now, 1)
            return True

        if count < self.limit:
            storage[key] = (last, count + 1)
            return True

        return False


class RedisRateLimiter(RateLimiter):
    """Redis rate limiter."""

    # TODO: Asyncio lock

    def __init__(self, limit: int, *, redis, **opts):
        """Initialize the rate limiter.

        Args:
            limit (int): The limit of requests.
            period (int): The period of time in seconds.
            redis (aioredis.Redis): The Redis connection.
        """
        super().__init__(limit, **opts)
        self.redis = redis

    async def check(self, key: str) -> bool:
        """Check the request."""
        value = await self.redis.get(key)
        if value is None:
            await self.redis.setex(key, self.period, 1)
            return True

        await self.redis.incr(key)
        return int(value) < self.limit



"""REST Utils."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Iterable, Mapping

if TYPE_CHECKING:
    from muffin import Request

    from muffin_rest.types import TVCollection


class Mutate(abc.ABC):
    """Mutate collections."""

    field: Any = None

    def __init__(self, name: str, *, field=None, **meta):
        """Initialize a name."""
        self.name = name
        self.field = field if field is not None else self.field or name
        self.meta = meta

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.name}'>"

    @abc.abstractmethod
    async def apply(self, collection: TVCollection) -> TVCollection:
        """Apply the mutation."""
        raise NotImplementedError


class Mutator(abc.ABC):
    """Mutate collections."""

    MUTATE_CLASS: type[Mutate]
    mutations: Mapping[str, Mutate]

    def __init__(self, handler, params: Iterable):
        """Initialize the mutations."""
        self.handler = handler
        self.mutations = {}
        for param in params:
            obj, meta = param if isinstance(param, tuple) else (param, {})
            mut = self.convert(obj, **meta)
            if mut:
                self.mutations[mut.name] = mut

    def __iter__(self):
        """Iterate through self filters."""
        return iter(self.mutations)

    def __str__(self) -> str:
        """Describe the filters."""
        return ",".join(self.mutations)

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self}'>"

    def __bool__(self):
        return bool(self.mutations)

    def convert(self, obj, **meta) -> Mutate:
        """Convert params to mutations."""
        if isinstance(obj, self.MUTATE_CLASS):
            return obj

        return self.MUTATE_CLASS(obj, **meta)

    @abc.abstractmethod
    async def apply(
        self, request: Request, collection: TVCollection
    ) -> tuple[TVCollection, dict[str, Any]]:
        """Mutate a collection."""
        raise NotImplementedError



"""Base class for API REST Handlers."""

import abc
import inspect
from typing import (
    Any,
    Generator,
    Generic,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Union,
    cast,
    overload,
)

import marshmallow as ma
from asgi_tools.response import ResponseJSON, parse_response
from muffin import Request
from muffin.handler import Handler, HandlerMeta

from muffin_rest import LIMIT_PARAM, OFFSET_PARAM, openapi
from muffin_rest.api import API
from muffin_rest.errors import APIError
from muffin_rest.filters import Filter
from muffin_rest.marshmallow import load_data
from muffin_rest.sorting import Sort
from muffin_rest.types import TSchemaRes

from .errors import HandlerNotBindedError
from .options import RESTOptions
from .types import TVCollection, TVData, TVResource


class RESTHandlerMeta(HandlerMeta):
    """Create class options."""

    def __new__(mcs, name, bases, params):
        """Prepare options for the handler."""
        kls = cast(type["RESTBase"], super().__new__(mcs, name, bases, params))
        kls.meta = kls.meta_class(kls)

        if getattr(kls.meta, kls.meta_class.base_property, None) is not None:
            kls.meta.filters = kls.meta.filters_cls(kls, kls.meta.filters)
            kls.meta.sorting = kls.meta.sorting_cls(kls, kls.meta.sorting)

        return kls


class RESTBase(Generic[TVResource], Handler, metaclass=RESTHandlerMeta):
    """Load/save resources."""

    auth: Any
    collection: Any
    resource: Any

    meta: RESTOptions
    meta_class: type[RESTOptions] = RESTOptions
    _api: Optional[API] = None

    filters: Optional[dict[str, Any]] = None
    sorting: Optional[dict[str, Any]] = None

    class Meta:
        """Tune the handler."""

        # Resource filters
        filters: Sequence[Union[str, tuple[str, str], Filter]] = ()

        # Define allowed resource sorting params
        sorting: Sequence[Union[str, tuple[str, dict], Sort]] = ()

        # Serialize/Deserialize Schema class
        Schema: Optional[type[ma.Schema]] = None

    @classmethod
    def __route__(cls, router, *paths, **params):
        """Bind the class to the given router."""
        methods = params.pop("methods") or cls.methods
        if paths:
            router.bind(cls, *paths, methods=methods, **params)

        else:
            router.bind(cls, f"/{ cls.meta.name }", methods=methods, **params)
            router.bind(
                cls, f"/{ cls.meta.name }/{{{ cls.meta.name_id }}}", methods=methods, **params
            )

        for _, method in inspect.getmembers(cls, lambda m: hasattr(m, "__route__")):
            paths, methods = method.__route__
            router.bind(cls, *paths, methods=methods, method_name=method.__name__)

        return cls

    async def __call__(self, request: Request, *, method_name: Optional[str] = None, **_) -> Any:
        """Dispatch the given request by HTTP method."""
        self.auth = await self.authorize(request)

        meta = self.meta

        self.collection = await self.prepare_collection(request)
        resource = await self.prepare_resource(request)
        method = getattr(self, method_name or request.method.lower())
        if not (request.method == "GET" and resource is None and not method_name):
            return await method(request, resource=resource)

        # Filter collection
        self.collection, self.filters = await self.filter(request, self.collection)

        # Sort collection
        self.collection, self.sorting = await self.sort(request, self.collection)

        # Paginate the collection
        headers = None
        if meta.limit:
            limit, offset = self.paginate_prepare_params(request)
            if limit and offset >= 0:
                self.collection, total = await self.paginate(request, limit=limit, offset=offset)
                headers = self.paginate_prepare_headers(limit, offset, total)

        response = await method(request)

        if headers:
            response = parse_response(response)
            response.headers.update(headers)

        return response

    @property
    def api(self) -> API:
        """Check if the handler is binded to an API."""
        if self._api is None:
            raise HandlerNotBindedError
        return self._api

    async def authorize(self, request: Request) -> Any:
        """Default authorization method. Proxy auth to self.api."""
        auth = await self.api.authorize(request)
        if not auth:
            raise APIError.UNAUTHORIZED()
        return auth

    # Prepare data
    # ------------
    @abc.abstractmethod
    async def prepare_collection(self, request: Request) -> Any:
        """Prepare a collection of resources. Create queryset, db cursor and etc."""
        raise NotImplementedError

    async def prepare_resource(self, request: Request) -> Any:
        """Load a resource."""
        return request["path_params"].get(self.meta.name_id)

    async def filter(self, request: Request, collection: TVCollection) -> tuple[TVCollection, Any]:
        """Filter the collection."""
        filters = self.meta.filters
        if filters:
            return await filters.apply(request, collection)

        return collection, None

    async def sort(self, request: Request, collection: TVCollection) -> tuple[TVCollection, Any]:
        sorting = self.meta.sorting
        if sorting:
            return await sorting.apply(request, collection)

        return collection, None

    # Paginate
    # --------
    def paginate_prepare_headers(self, limit, offset, total=None):
        """Prepare pagination headers."""
        headers = {"x-limit": limit, "x-offset": offset}
        if total is not None:
            headers["x-total"] = total
        return headers

    def paginate_prepare_params(self, request: Request) -> tuple[int, int]:
        """Prepare pagination params."""
        meta = self.meta
        query = request.url.query
        limit = query.get(LIMIT_PARAM) or meta.limit
        try:
            return min(abs(int(limit)), meta.limit_max), int(query.get(OFFSET_PARAM, 0))
        except ValueError as exc:
            raise APIError.BAD_REQUEST("Pagination params are invalid") from exc

    @abc.abstractmethod
    async def paginate(
        self, request: Request, *, limit: int = 0, offset: int = 0
    ) -> tuple[Any, Optional[int]]:
        """Paginate the results."""
        raise NotImplementedError

    # Manage data
    # -----------
    @abc.abstractmethod
    async def save(self, request: Request, resource: TVResource, *, update=False) -> TVResource:
        """Save the given resource."""
        return resource

    async def save_many(
        self, request: Request, data: list[TVResource], *, update=False
    ) -> list[TVResource]:
        """Save many resources."""
        return [await self.save(request, item, update=update) for item in data]

    @abc.abstractmethod
    async def remove(self, request: Request, resource):
        """Remove the given resource."""
        raise NotImplementedError

    # Parse data
    # -----------
    def get_schema(
        self, request: Request, *, resource: Optional[TVResource] = None, **schema_options
    ) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        query = request.url.query
        schema_options.setdefault("only", query.get("schema_only"))
        schema_options.setdefault("exclude", query.get("schema_exclude", ()))
        return self.meta.Schema(**schema_options)

    async def load_data(self, request: Request):
        """Load data from request and create/update a resource."""
        try:
            data = await request.data(raise_errors=True)
        except (ValueError, TypeError) as err:
            raise APIError.BAD_REQUEST(str(err)) from err

        return data

    async def load(
        self, request: Request, resource: Optional[TVResource] = None, **schema_options
    ) -> TVData[TVResource]:
        """Load data from request and create/update a resource."""
        schema = self.get_schema(request, resource=resource, **schema_options)
        data = cast(Union[Mapping, list], await self.load_data(request))
        return cast(TVData[TVResource], await load_data(data, schema, partial=resource is not None))

    @overload
    async def dump(  # type: ignore[misc]
        self, request, data: TVData, *, many: Literal[True]
    ) -> list[TSchemaRes]: ...

    @overload
    async def dump(self, request, data: TVData, *, many: bool = False) -> TSchemaRes: ...

    async def dump(
        self,
        request: Request,
        data: Union[TVResource, Iterable[TVResource]],
        *,
        many: bool = False,
    ) -> Union[TSchemaRes, list[TSchemaRes]]:
        """Serialize the given response."""
        schema = self.get_schema(request)
        return schema.dump(data, many=many)

    async def get(self, request: Request, *, resource: Optional[TVResource] = None) -> ResponseJSON:
        """Get a resource or a collection of resources.

        Specify a path param to load a resource.
        """
        res = await (
            self.dump(request, resource)
            if resource
            else self.dump(request, data=self.collection, many=True)
        )
        return ResponseJSON(res)

    async def post(
        self, request: Request, *, resource: Optional[TVResource] = None
    ) -> ResponseJSON:
        """Create a resource.

        The method accepts a single resource's data or a list of resources to create.
        """
        data = await self.load(request, resource)
        many = isinstance(data, list)
        if many:
            data = await self.save_many(request, data, update=resource is not None)
        else:
            data = await self.save(request, cast(TVResource, data), update=resource is not None)

        res = await self.dump(request, data, many=many)
        return ResponseJSON(res)

    async def put(self, request: Request, *, resource: Optional[TVResource] = None) -> ResponseJSON:
        """Update a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        return await self.post(request, resource=resource)

    async def delete(self, request: Request, resource: Optional[TVResource] = None):
        """Delete a resource."""
        if resource is None:
            raise APIError.NOT_FOUND()

        res = await self.remove(request, resource)
        return ResponseJSON(res)


class RESTHandler(RESTBase[TVResource], openapi.OpenAPIMixin):
    """Basic Handler Class."""


def to_sort(sort_params: Sequence[str]) -> Generator[tuple[str, bool], None, None]:
    """Generate sort params."""
    for name in sort_params:
        n, desc = name.strip("-"), name.startswith("-")
        if n:
            yield n, desc



"""Mongo Utils."""

from __future__ import annotations

from typing import TYPE_CHECKING, Awaitable, Union

if TYPE_CHECKING:
    from motor import motor_asyncio as motor


class MongoChain:
    """Support query chains.

    Only for `find` and `find_one` methods.
    ::
        collection = MongoChain(mongo_collection)
        collection = collection.find({'field': 'value').find('field2': 'value')
        result = collection.find_one({'field3': 'value')
        results = collection.skip(10).limit(10)
    """

    CURSOR_METHODS = (
        "where",
        "sort",
        "skip",
        "rewind",
        "retrieved",
        "remove_option",
        "next",
        "min",
        "max_time_ms",
        "max_scan",
        "max_await_time_ms",
        "max",
        "limit",
        "hint",
        "explain",
        "distinct",
        "cursor_id",
        "count",
        "comment",
        "collection",
        "close",
        "clone",
        "batch_size",
        "alive",
        "address",
        "add_option",
        "__getitem__",
    )

    def __init__(self, collection: motor.AsyncIOMotorCollection):
        """Initialize the resource."""
        self.collection = collection
        self.query: list = []
        self.projection = None
        self.sorting: list[tuple[str, int]] = []

    def find(
        self,
        query: Union[list, dict, None] = None,
        projection=None,
    ) -> MongoChain:
        """Store filters in self."""
        self.query = self.__update__(query)
        self.projection = projection
        return self

    def find_one(
        self,
        query: Union[list, dict, None] = None,
        projection=None,
    ) -> Awaitable:
        """Apply filters and return cursor."""
        query = self.__update__(query)
        query = query and {"$and": query} or {}
        return self.collection.find_one(query, projection=projection)

    def count(self) -> Awaitable[int]:
        """Count documents."""
        query = self.query and {"$and": self.query} or {}
        return self.collection.count_documents(query)

    def aggregate(self, pipeline, **kwargs):
        """Aggregate collection."""
        if self.query:
            for params in pipeline:
                if "$match" in params:
                    query = self.__update__(params["$match"])
                    params["$match"] = {"$and": query}
                    break
            else:
                pipeline.insert(0, {"$match": {"$and": self.query}})

        if self.sorting:
            pipeline = [p for p in pipeline if "$sort" not in p]
            pipeline.append({"$sort": dict(self.sorting)})

        return self.collection.aggregate(pipeline, **kwargs)

    def sort(self, key, direction=1):
        """Save ordering properties."""
        if isinstance(key, str):
            self.sorting.append((key, direction))

        else:
            self.sorting = key

        return self

    def __repr__(self):
        """String representation."""
        return "<MongoChain (%s) %r>" % (self.collection.name, self.query)

    def __update__(self, query):
        """Update stored query."""
        if query:
            self.query.append(query)

        return self.query

    def __iter__(self):
        """Iterate by self collection."""
        query = self.query and {"$and": self.query} or {}
        if self.sorting:
            return self.collection.find(query, self.projection).sort(self.sorting)

        return self.collection.find(query, self.projection)

    def __getattr__(self, name):
        """Proxy any attributes except find to self.collection."""
        if name in self.CURSOR_METHODS:
            query = self.query and {"$and": self.query} or {}
            cursor = self.collection.find(query, self.projection)
            if self.sorting:
                cursor = cursor.sort(self.sorting)
            return getattr(cursor, name)
        return getattr(self.collection, name)



from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from .handler import RESTBase

from muffin import Request

TVCollection = TypeVar("TVCollection", bound=Any)
TVResource = TypeVar("TVResource", bound=Any)
TVData = Union[TVResource, list[TVResource]]
TAuth = Callable[[Request], Awaitable]
TVAuth = TypeVar("TVAuth", bound=TAuth)
TVHandler = TypeVar("TVHandler", bound=type["RESTBase"])
TSchemaRes = dict[str, Any]

TFilterValue = tuple[Callable, Any]
TFilterOps = tuple[TFilterValue, ...]



"""SQLAlchemy Core support."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, cast

import marshmallow as ma
import sqlalchemy as sa
from marshmallow_sqlalchemy import ModelConverter
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema as BaseSQLAlchemyAutoSchema

from muffin_rest.errors import APIError
from muffin_rest.handler import RESTHandler, RESTOptions
from muffin_rest.sqlalchemy.filters import SAFilters
from muffin_rest.sqlalchemy.sorting import SASorting

if TYPE_CHECKING:
    from muffin import Request
    from muffin_databases import Plugin as Database

    from muffin_rest.types import TVData

from .types import TVResource

ModelConverter._get_field_name = lambda _, prop_or_column: str(prop_or_column.key)  # type: ignore[method-assign]


class SQLAlchemyAutoSchema(BaseSQLAlchemyAutoSchema):
    """Allow partial updates for tables."""

    @ma.pre_load
    def fill_defaults(self, data, *, partial=False, **_):
        """Insert default params for SQLAlchemy because databases don't.

        https://github.com/encode/databases/issues/72
        """
        cols_to_fields = {f.attribute or f.name: f for f in self.declared_fields.values()}
        if not partial:
            for column in self.opts.table.columns:
                field = cols_to_fields.get(column.name)
                if not field:
                    continue

                data_key = field.data_key or field.name
                if data_key not in data and column.default is not None:
                    value = column.default.arg
                    if callable(value):
                        value = value(column)

                    data[data_key] = field._serialize(
                        value,
                        field.attribute or field.name,
                        None,
                    )

        return data

    @ma.post_load
    def make_instance(self, data, **kwargs):
        """Update a table instance."""
        if self.instance is not None:
            instance = dict(self.instance)
            instance.update(data)
            return instance
        return super().make_instance(data, **kwargs)


class SARESTOptions(RESTOptions):
    """Support SQLAlchemy Core."""

    filters_cls: type[SAFilters] = SAFilters
    sorting_cls: type[SASorting] = SASorting

    # Schema auto generation params
    Schema: type[SQLAlchemyAutoSchema]
    schema_base: type[SQLAlchemyAutoSchema] = SQLAlchemyAutoSchema

    table: sa.Table
    table_pk: sa.Column
    database: Database

    base_property = "table"

    def setup(self, cls):
        """Prepare meta options."""
        if self.database is None:
            raise ValueError("'SARESTHandler.Meta.database' is required")

        self.name = self.name or self.table.name
        self.table_pk = getattr(self, "table_pk", None) or self.table.c.id

        super().setup(cls)

    def setup_schema_meta(self, _):
        """Prepare a schema."""
        return type(
            "Meta",
            (object,),
            dict(
                {
                    "unknown": self.schema_unknown,
                    "table": self.table,
                    "include_fk": True,
                    "dump_only": (self.name_id,),
                },
                **self.schema_meta,
            ),
        )


class SARESTHandler(RESTHandler[TVResource]):
    """Support SQLAlchemy Core."""

    meta: SARESTOptions
    meta_class: type[SARESTOptions] = SARESTOptions
    collection: sa.sql.Select

    async def prepare_collection(self, _: Request) -> sa.sql.Select:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.table.select()

    async def paginate(
        self,
        _: Request,
        *,
        limit: int = 0,
        offset: int = 0,
    ) -> tuple[sa.sql.Select, Optional[int]]:
        """Paginate the collection."""
        sqs = self.collection.order_by(None).subquery()
        qs = sa.select(sa.func.count()).select_from(sqs)
        total = None
        if self.meta.limit_total:
            total = await self.meta.database.fetch_val(qs)
        return self.collection.offset(offset).limit(limit), total

    async def get(self, request, *, resource: Optional[TVResource] = None) -> Any:
        """Get resource or collection of resources."""
        if resource:
            return await self.dump(request, resource)

        rows = await self.meta.database.fetch_all(self.collection)
        return await self.dump(request, rows, many=True)

    async def prepare_resource(self, request: Request) -> Optional[TVResource]:
        """Load a resource."""
        pk = request["path_params"].get(self.meta.name_id)
        if not pk:
            return None

        qs = self.collection.where(self.meta.table_pk == pk)
        resource = await self.meta.database.fetch_one(qs)
        if resource is None:
            raise APIError.NOT_FOUND("Resource not found")
        return cast(TVResource, dict(resource))

    def get_schema(
        self, request: Request, *, resource: Optional[TVResource] = None, **schema_options
    ) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return super().get_schema(request, instance=resource, **schema_options)

    async def save(self, _: Request, resource: TVData[TVResource], *, update=False):
        """Save the given resource."""
        meta = self.meta
        insert_query = meta.table.insert()
        table_pk = cast(sa.Column, meta.table_pk)
        if update:
            update_query = self.meta.table.update().where(table_pk == resource[table_pk.name])  # type: ignore[call-overload]
            await meta.database.execute(update_query, resource)

        else:
            resource[table_pk.name] = await meta.database.execute(insert_query, resource)  # type: ignore[call-overload]

        return resource

    async def remove(self, request: Request, resource: Optional[TVResource] = None):
        """Remove the given resource."""
        table_pk = cast(sa.Column, self.meta.table_pk)
        pks = [resource[table_pk.name]] if resource else await request.data()
        if not pks:
            raise APIError.NOT_FOUND()

        delete = self.meta.table.delete().where(table_pk.in_(cast(list[Any], pks)))
        await self.meta.database.execute(delete)

    delete = remove



import marshmallow as ma
from marshmallow_peewee import DefaultConverter
from muffin_peewee.fields import IntEnumField, StrEnumField, URLField

from muffin_rest.schemas import EnumField


@DefaultConverter.register(StrEnumField)
@DefaultConverter.register(IntEnumField)
def build_field(field, opts, **params):
    params.pop("validate", None)
    return EnumField(field.enum, **params)


DefaultConverter.register(URLField, ma.fields.Url)

# ruff: noqa: ARG001, ARG002



"""Support filters for Mongo."""

from typing import ClassVar

from muffin_rest.filters import Filter, Filters
from muffin_rest.types import TFilterValue


class MongoFilter(Filter):
    """Custom filter for sqlalchemy."""

    operators: ClassVar = {
        "$eq": lambda _, v: ("$eq", v),
        "$ge": lambda _, v: ("$ge", v),
        "$gt": lambda _, v: ("$gt", v),
        "$in": lambda _, v: ("$in", v),
        "$le": lambda _, v: ("$le", v),
        "$lt": lambda _, v: ("$lt", v),
        "$ne": lambda _, v: ("$ne", v),
        "$nin": lambda _, v: ("$nin", v),
        "$starts": lambda _, v: ("$regex", f"^{ v }"),
        "$ends": lambda _, v: ("$regex", f"{ v }$"),
    }

    async def filter(self, collection, *ops: TFilterValue):
        """Apply the filter."""
        return collection.find({self.field: dict(op(self.name, v) for op, v in ops)})


class MongoFilters(Filters):
    """Bind MongoFilter class."""

    MUTATE_CLASS = MongoFilter



from typing import Any, TypeVar

from .utils import MongoChain

TResource = dict[str, Any]
TVCollection = TypeVar("TVCollection", bound=MongoChain)
TVResource = TypeVar("TVResource", bound=TResource)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from marshmallow import Schema, ValidationError

from muffin_rest.errors import APIError

if TYPE_CHECKING:
    from collections.abc import Mapping


async def load_data(data: Union[Mapping, list], schema: Optional[Schema] = None, **params):
    if schema is None:
        return data

    try:
        return schema.load(data, many=isinstance(data, list), **params)
    except ValidationError as err:
        raise APIError.BAD_REQUEST("Bad request data", errors=err.messages) from err
